import os

from symsynd.mach import get_cpu_name, get_macho_uuids


def find_debug_images(dsym_path, binary_images):
    base = os.path.join(dsym_path, 'Contents', 'Resources', 'DWARF')

    images_to_load = set()

    for image in binary_images:
        cpu_name = get_cpu_name(image['cpu_type'],
                                image['cpu_subtype'])
        if cpu_name is not None:
            images_to_load.add(image['uuid'].lower())

    images = {}

    # Step one: load images that are named by their UUID
    for uuid in list(images_to_load):
        fn = os.path.join(dsym_path, uuid)
        if os.path.isfile(fn):
            images[uuid] = fn
            images_to_load.discard(uuid)

    # Otherwise fall back to loading images from the dsym bundle.
    if images_to_load and os.path.isdir(base):
        for fn in os.listdir(base):
            # Looks like a UUID we loaded, skip it
            if fn in images:
                continue
            full_fn = os.path.join(base, fn)
            uuids = get_macho_uuids(full_fn)
            for _, uuid in uuids:
                if uuid in images_to_load:
                    images[uuid] = full_fn
                    images_to_load.discard(uuid)

    rv = {}

    # Now resolve all the images.
    for image in binary_images:
        cpu_name = get_cpu_name(image['cpu_type'],
                                image['cpu_subtype'])
        if cpu_name is None:
            continue
        uid = image['uuid'].lower()
        if uid not in images:
            continue
        rv[image['image_addr']] = {
            'uuid': uid,
            'image_addr': image['image_addr'],
            'dsym_path': images[uid],
            'image_vmaddr': image['image_vmaddr'],
            'cpu_name': cpu_name,
        }

    return rv


class ReportSymbolizer(object):

    def __init__(self, driver, dsym_path, binary_images):
        self.driver = driver
        self.images = find_debug_images(dsym_path, binary_images)

    def symbolize_backtrace(self, backtrace):
        rv = []
        for frame in backtrace:
            frame = dict(frame)
            img = self.images.get(frame['object_addr'])
            if img is not None:
                frame.update(self.driver.symbolize(
                    img['dsym_path'], img['image_vmaddr'],
                    img['image_addr'], frame['instruction_addr'],
                    img['cpu_name']))
            rv.append(frame)
        return rv



import subprocess
from cffi import FFI

ffi = FFI()
ffi.cdef(subprocess.Popen([
    'cc', '-E', '-DDM_NOINCLUDE',
    'demangle-swift/demangle-swift.h'],
    stdout=subprocess.PIPE).communicate()[0])

with open('demangle-swift/demangle-all.cpp') as source:
    ffi.set_source(
        '_symsynd_swift_demangler',
        source.read(),
        include_dirs=['demangle-swift'],
        extra_compile_args=['-std=c++1y'],
        source_extension='.cpp'
    )



import uuid
from macholib.MachO import MachO
from macholib.mach_o import uuid_command


CPU_ARCH_MASK = 0xff000000
CPU_ARCH_ABI64 = 0x01000000
CPU_TYPE_ANY = -1
CPU_TYPE_VAX = 1
CPU_TYPE_MC680x0 = 6
CPU_TYPE_X86 = 7
CPU_TYPE_I386 = CPU_TYPE_X86
CPU_TYPE_X86_64 = CPU_TYPE_X86 | CPU_ARCH_ABI64
CPU_TYPE_MC98000 = 10
CPU_TYPE_HPPA = 11
CPU_TYPE_ARM = 12
CPU_TYPE_ARM64 = CPU_TYPE_ARM | CPU_ARCH_ABI64
CPU_TYPE_MC88000 = 13
CPU_TYPE_SPARC = 14
CPU_TYPE_I860 = 15
CPU_TYPE_POWERPC = 18
CPU_TYPE_POWERPC64 = CPU_TYPE_POWERPC | CPU_ARCH_ABI64
CPU_SUBTYPE_MASK = 0xff000000
CPU_SUBTYPE_LIB64 = 0x80000000
CPU_SUBTYPE_MULTIPLE = -1
CPU_SUBTYPE_LITTLE_ENDIAN = 0
CPU_SUBTYPE_BIG_ENDIAN = 1
CPU_THREADTYPE_NONE = 0
CPU_SUBTYPE_VAX_ALL = 0
CPU_SUBTYPE_VAX780 = 1
CPU_SUBTYPE_VAX785 = 2
CPU_SUBTYPE_VAX750 = 3
CPU_SUBTYPE_VAX730 = 4
CPU_SUBTYPE_UVAXI = 5
CPU_SUBTYPE_UVAXII = 6
CPU_SUBTYPE_VAX8200 = 7
CPU_SUBTYPE_VAX8500 = 8
CPU_SUBTYPE_VAX8600 = 9
CPU_SUBTYPE_VAX8650 = 10
CPU_SUBTYPE_VAX8800 = 11
CPU_SUBTYPE_UVAXIII = 12

CPU_SUBTYPE_MC680x0_ALL = 1
CPU_SUBTYPE_MC68030 = 1
CPU_SUBTYPE_MC68040 = 2
CPU_SUBTYPE_MC68030_ONLY = 3
CPU_SUBTYPE_INTEL = lambda f, m: (f) + ((m) << 4)
CPU_SUBTYPE_I386_ALL = CPU_SUBTYPE_INTEL(3, 0)
CPU_SUBTYPE_386 = CPU_SUBTYPE_INTEL(3, 0)
CPU_SUBTYPE_486 = CPU_SUBTYPE_INTEL(4, 0)
CPU_SUBTYPE_486SX = CPU_SUBTYPE_INTEL(4, 8)
CPU_SUBTYPE_586 = CPU_SUBTYPE_INTEL(5, 0)
CPU_SUBTYPE_PENT = CPU_SUBTYPE_INTEL(5, 0)
CPU_SUBTYPE_PENTPRO = CPU_SUBTYPE_INTEL(6, 1)
CPU_SUBTYPE_PENTII_M3 = CPU_SUBTYPE_INTEL(6, 3)
CPU_SUBTYPE_PENTII_M5 = CPU_SUBTYPE_INTEL(6, 5)
CPU_SUBTYPE_CELERON = CPU_SUBTYPE_INTEL(7, 6)
CPU_SUBTYPE_CELERON_MOBILE = CPU_SUBTYPE_INTEL(7, 7)
CPU_SUBTYPE_PENTIUM_3 = CPU_SUBTYPE_INTEL(8, 0)
CPU_SUBTYPE_PENTIUM_3_M = CPU_SUBTYPE_INTEL(8, 1)
CPU_SUBTYPE_PENTIUM_3_XEON = CPU_SUBTYPE_INTEL(8, 2)
CPU_SUBTYPE_PENTIUM_M = CPU_SUBTYPE_INTEL(9, 0)
CPU_SUBTYPE_PENTIUM_4 = CPU_SUBTYPE_INTEL(10, 0)
CPU_SUBTYPE_PENTIUM_4_M = CPU_SUBTYPE_INTEL(10, 1)
CPU_SUBTYPE_ITANIUM = CPU_SUBTYPE_INTEL(11, 0)
CPU_SUBTYPE_ITANIUM_2 = CPU_SUBTYPE_INTEL(11, 1)
CPU_SUBTYPE_XEON = CPU_SUBTYPE_INTEL(12, 0)
CPU_SUBTYPE_XEON_MP = CPU_SUBTYPE_INTEL(12, 1)
CPU_SUBTYPE_INTEL_FAMILY = lambda x: (x) & 15
CPU_SUBTYPE_INTEL_FAMILY_MAX = 15
CPU_SUBTYPE_INTEL_MODEL = lambda x: (x) >> 4
CPU_SUBTYPE_INTEL_MODEL_ALL = 0
CPU_SUBTYPE_X86_ALL = 3
CPU_SUBTYPE_X86_64_ALL = 3
CPU_SUBTYPE_X86_ARCH1 = 4
CPU_THREADTYPE_INTEL_HTT = 1
CPU_SUBTYPE_MIPS_ALL = 0
CPU_SUBTYPE_MIPS_R2300 = 1
CPU_SUBTYPE_MIPS_R2600 = 2
CPU_SUBTYPE_MIPS_R2800 = 3
CPU_SUBTYPE_MIPS_R2000a = 4
CPU_SUBTYPE_MIPS_R2000 = 5
CPU_SUBTYPE_MIPS_R3000a = 6
CPU_SUBTYPE_MIPS_R3000 = 7
CPU_SUBTYPE_MC98000_ALL = 0
CPU_SUBTYPE_MC98601 = 1
CPU_SUBTYPE_HPPA_ALL = 0
CPU_SUBTYPE_HPPA_7100 = 0
CPU_SUBTYPE_HPPA_7100LC = 1
CPU_SUBTYPE_POWERPC_ALL = 0
CPU_SUBTYPE_POWERPC64_ALL = 0
CPU_SUBTYPE_POWERPC_601 = 1
CPU_SUBTYPE_POWERPC_602 = 2
CPU_SUBTYPE_POWERPC_603 = 3
CPU_SUBTYPE_POWERPC_603e = 4
CPU_SUBTYPE_POWERPC_603ev = 5
CPU_SUBTYPE_POWERPC_604 = 6
CPU_SUBTYPE_POWERPC_604e = 7
CPU_SUBTYPE_POWERPC_620 = 8
CPU_SUBTYPE_POWERPC_750 = 9
CPU_SUBTYPE_POWERPC_7400 = 10
CPU_SUBTYPE_POWERPC_7450 = 11
CPU_SUBTYPE_POWERPC_970 = 100
CPU_SUBTYPE_ARM_ALL = 0
CPU_SUBTYPE_ARM_V4T = 5
CPU_SUBTYPE_ARM_V6 = 6
CPU_SUBTYPE_ARM_V5TEJ = 7
CPU_SUBTYPE_ARM_XSCALE = 8
CPU_SUBTYPE_ARM_V7 = 9
CPU_SUBTYPE_ARM_V7F = 10
CPU_SUBTYPE_ARM_V7S = 11
CPU_SUBTYPE_ARM_V7K = 12
CPU_SUBTYPE_ARM_V6M = 14
CPU_SUBTYPE_ARM_V7M = 15
CPU_SUBTYPE_ARM_V7EM = 16
CPU_SUBTYPE_ARM_V8 = 13
CPU_SUBTYPE_ARM64_ALL = 0
CPU_SUBTYPE_ARM64_V8 = 1
CPUFAMILY_UNKNOWN = 0
CPUFAMILY_POWERPC_G3 = 0xcee41549
CPUFAMILY_POWERPC_G4 = 0x77c184ae
CPUFAMILY_POWERPC_G5 = 0xed76d8aa
CPUFAMILY_INTEL_6_13 = 0xaa33392b
CPUFAMILY_INTEL_6_14 = 0x73d67300
CPUFAMILY_INTEL_6_15 = 0x426f69ef
CPUFAMILY_INTEL_6_23 = 0x78ea4fbc
CPUFAMILY_INTEL_6_26 = 0x6b5a4cd2
CPUFAMILY_ARM_9 = 0xe73283ae
CPUFAMILY_ARM_11 = 0x8ff620d8
CPUFAMILY_ARM_XSCALE = 0x53b005f5
CPUFAMILY_ARM_13 = 0x0cc90e64
CPUFAMILY_INTEL_YONAH = CPUFAMILY_INTEL_6_14
CPUFAMILY_INTEL_MEROM = CPUFAMILY_INTEL_6_15
CPUFAMILY_INTEL_PENRYN = CPUFAMILY_INTEL_6_23
CPUFAMILY_INTEL_NEHALEM = CPUFAMILY_INTEL_6_26
CPUFAMILY_INTEL_CORE = CPUFAMILY_INTEL_6_14
CPUFAMILY_INTEL_CORE2 = CPUFAMILY_INTEL_6_15


CPU_NAMES = {
    (CPU_TYPE_X86_64, CPU_SUBTYPE_X86_64_ALL): 'x86_64',
    (CPU_TYPE_I386, CPU_SUBTYPE_I386_ALL): 'i386',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_ALL): 'arm',
    (CPU_TYPE_ARM64, CPU_SUBTYPE_ARM64_ALL): 'arm64',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_ALL): 'ppc',
    (CPU_TYPE_POWERPC64, CPU_SUBTYPE_POWERPC64_ALL): 'ppc64',
    (CPU_TYPE_MC680x0, CPU_SUBTYPE_MC680x0_ALL): 'm68k',
    (CPU_TYPE_HPPA, CPU_SUBTYPE_HPPA_ALL): 'hppa',
    (CPU_TYPE_I386, CPU_SUBTYPE_486): 'i486',
    (CPU_TYPE_I386, CPU_SUBTYPE_486SX): 'i486SX',
    (CPU_TYPE_I386, CPU_SUBTYPE_PENT): 'pentium',
    (CPU_TYPE_I386, CPU_SUBTYPE_586): 'i586',
    (CPU_TYPE_I386, CPU_SUBTYPE_PENTPRO): 'pentpro',
    (CPU_TYPE_I386, CPU_SUBTYPE_PENTPRO): 'i686',
    (CPU_TYPE_I386, CPU_SUBTYPE_PENTII_M3): 'pentIIm3',
    (CPU_TYPE_I386, CPU_SUBTYPE_PENTII_M5): 'pentIIm5',
    (CPU_TYPE_I386, CPU_SUBTYPE_PENTIUM_4): 'pentium4',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V4T): 'armv4t',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V5TEJ): 'armv5',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_XSCALE): 'xscale',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V6): 'armv6',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V6M): 'armv6m',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V7): 'armv7',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V7F): 'armv7f',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V7S): 'armv7s',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V7K): 'armv7k',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V7M): 'armv7m',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V7EM): 'armv7em',
    (CPU_TYPE_ARM, CPU_SUBTYPE_ARM_V8): 'armv8',
    (CPU_TYPE_ARM64, CPU_SUBTYPE_ARM64_V8): 'arm64',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_601): 'ppc601',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_603): 'ppc603',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_604): 'ppc604',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_604e): 'ppc604e',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_750): 'ppc750',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_7400): 'ppc7400',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_7450): 'ppc7450',
    (CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_970): 'ppc970',
    (CPU_TYPE_MC680x0, CPU_SUBTYPE_MC68030_ONLY): 'm68030',
    (CPU_TYPE_MC680x0, CPU_SUBTYPE_MC68040): 'm68040',
    (CPU_TYPE_HPPA, CPU_SUBTYPE_HPPA_7100LC): 'hppa7100LC',
}

VALID_CPU_NAMES = frozenset(CPU_NAMES.values())


def get_cpu_name(type, subtype):
    """Get a known CPU name from the given maco cpu type and subtype.
    If this is not known `None` is returned.  This only covers things we
    actually expect in OS X/iOS projects.
    """
    return CPU_NAMES.get((type, subtype))


def is_valid_cpu_name(name):
    return name in VALID_CPU_NAMES


def get_macho_uuids(filename):
    """Given the filename to a mach-o file this returns the UUIDs for the
    binaries within it.
    """
    rv = []
    bin = MachO(filename)
    for header in bin.headers:
        cpu = get_cpu_name(header.header.cputype,
                           header.header.cpusubtype)
        if cpu is None:
            continue
        for cmd in header.commands:
            if type(cmd[1]) is uuid_command:
                rv.append((cpu, str(uuid.UUID(bytes=cmd[1].uuid))))
    return rv



from _symsynd_swift_demangler import ffi, lib


def demangle_symbol(symbol):
    buffer = ffi.new('char[4000]')
    rv = lib.demangle_swift(symbol, buffer, 4000)
    if rv:
        return ffi.string(buffer)



import os
from itertools import chain


def which(prog, extra_paths=None):
    path = os.environ['PATH'].split(os.path.pathsep)
    if extra_paths:
        path = chain(path, extra_paths)
    for p in path:
        p = os.path.join(p, prog)
        if os.path.exists(p) and os.access(p, os.X_OK):
            return p



import os
from setuptools import setup, find_packages


# Build with clang if not otherwise specified.
os.environ.setdefault('CC', 'clang')


setup(
    name='symsynd',
    version='0.1.3',
    url='http://github.com/getsentry/symsynd',
    description='Helps symbolicating crash dumps.',
    license='BSD',
    author='Sentry',
    author_email='hello@getsentry.com',
    packages=find_packages(),
    cffi_modules=['demangler_build.py:ffi'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'macholib',
        'cffi>=1.0.0',
    ],
    setup_requires=[
        'cffi>=1.0.0'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



import os
import sys
import errno
import subprocess
from threading import Lock

from symsynd.utils import which
from symsynd.mach import is_valid_cpu_name
from symsynd.swift import demangle_symbol as demangle_swift_symbol


devnull = open(os.path.devnull, 'a')


SYMBOLIZER_SEARCHPATHS = []
if sys.platform == 'darwin':
    SYMBOLIZER_SEARCHPATHS.append('/usr/local/opt/llvm/bin')


def qm_to_none(value):
    if value == '??':
        return None
    return value


def normalize_dsym_path(p):
    if '\x00' in p or '"' in p or '\n' in p or '\r' in p:
        raise ValueError('Invalid character in dsym path')
    p = os.path.abspath(p)
    if not os.path.isfile(p):
        raise IOError(errno.ENOENT, 'dsym file not found (%s)' % p)
    return p


def find_llvm_symbolizer():
    p = os.environ.get('LLVM_SYMBOLIZER_PATH')
    if p:
        return p

    p = which('llvm-symbolizer', extra_paths=SYMBOLIZER_SEARCHPATHS)
    if p is not None:
        return p

    for ver in xrange(4, 12):
        p = which('llvm-symbolizer-3.%d' % ver)
        if p is not None:
            return p


class Driver(object):

    def __init__(self, symbolizer_path=None):
        if symbolizer_path is None:
            symbolizer_path = find_llvm_symbolizer()
        self._proc = subprocess.Popen([symbolizer_path],
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE,
                                      stderr=devnull)
        self._lock = Lock()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def close(self):
        if self._proc is not None:
            self._proc.kill()
            self._proc = None

    def symbolize(self, dsym_path, image_vmaddr, image_addr,
                  instruction_addr, cpu_name):
        if self._proc is None:
            raise RuntimeError('Symbolizer is closed')
        if not is_valid_cpu_name(cpu_name):
            raise ValueError('"%s" is not a valid cpu name' % cpu_name)
        dsym_path = normalize_dsym_path(dsym_path)

        addr = image_vmaddr + instruction_addr - image_addr
        stdout = self._proc.stdout
        with self._lock:
            self._proc.stdin.write('"%s:%s" 0x%x\n' % (
                dsym_path,
                cpu_name,
                addr,
            ))
            sym = stdout.readline().rstrip()
            location = stdout.readline().rstrip()
            stdout.readline()

        pieces = location.rsplit(':', 3)
        sym = qm_to_none(sym)

        if sym is not None:
            sym = (demangle_swift_symbol(sym) or sym).decode('utf-8', 'replace')

        return {
            'symbol_name': sym,
            'filename': qm_to_none(pieces[0].decode('utf-8')),
            'line': int(pieces[1]),
            'column': int(pieces[2]),
        }
